/*
  Babix - An educational little preemptive multitask kernel for Arduino Due.
  Fran�ois Pessaux 04/2015.
  Maxime Ayrault 06/2015.
  This code can be freely distributed.
*/

#include <Arduino.h>
#include "semaphore.h"


/***
    Attempt to acquire one of the lock. This function used the number of the
    lock you are going to use as argument
 ***/
void sem_acquire (semaphore_t *sem)
{
  /*  No need to save the registers used in this function. In effect, we only
      use scratch registers r1 and r2. Gcc seems to store our argument in r0
      which is also a scratch register. */

  /* The label. */
 sem_retry:
  /* Ask an exclusive access on the lock and get its value by the way
     (c.f. datasheet section 12.5.7 page 78 and section 12.12.8 page 112.  */
  asm volatile ("ldrex r1, [%0]   \n\t" :: "r" (sem) : "r1", "memory") ;

  /* Check the value of the lock. If it is full (==0) we must retry.  */
  asm volatile ("mov r2, #0  \n\t" ::: "r2") ;
  asm volatile ("cmp r2, r1  \n\t") ;
  /* Attention: must use an asm goto to make sure that gcc will know that
     we need to go back and re exec the previous code, to avoid register
     problem.
     If you are not going to use that, gcc will not kown that a label exist,
     and  you can kill the register with the lock address.
  */
  asm volatile goto ("beq %0   \n\t" :::: sem_retry) ;  /* Loop back. */

  /* Lock is not == 0, hence is not full. Let's try to acquire it and
     decrement it's value. */
  asm volatile ("sub r1, r1, #1  \n\t" ::: "r1") ;
  /* And stock it in the right lock */
  asm volatile ("strex r2, r1, [%0]  \n\t" :: "r" (sem) : "r2", "memory") ;

  /*  Check return value: if 0 write succeeded otherwise failure. In case of
      failure we must retry the whole process (i.e. reading the lock's value
      and if possible trying to write 1 inside. */
  asm volatile ("mov r1, #1  \n\t" ::: "r1") ;
  asm volatile ("cmp r2, r1      \n\t") ;
  asm volatile goto ("beq %0   \n\t" :::: sem_retry) ;  /* Loop back. */
  return ;
}



/***
    Release one lock whose number is passed as argument.
    Typically, this is to call after the critical section.
***/
void sem_release (semaphore_t *sem)
{
 rel_retry:
  asm volatile ("ldrex r1, [%0]   \n\t" :: "r" (sem) : "r1", "memory") ;
  asm volatile ("add r1, r1, #1\n\t") ;
  asm volatile ("strex r2, r1, [%0]  \n\t" :: "r" (sem) : "r2", "memory") ;
  asm volatile ("mov r1, #1  \n\t" ::: "r1") ;
  asm volatile ("cmp r2, r1      \n\t") ;
  asm volatile goto ("beq %0   \n\t" :::: rel_retry) ;  /* Loop back. */
  return ;
}
